==> pvfs-1.6.1-pre1.tgz: pvfs-1.6.1-pre1/lib/pvfs_open.c <==
/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See LIBRARY_COPYING in top-level directory.
 */


/* This file contains PVFS library call for open.   		*/
/* Should work to create and open files 						*/

#include <lib.h>
#include <signal.h>
#include <sockio.h>
#include <sockset.h>
#include <stdarg.h>
#include <errno.h>

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

extern sockset socks;
extern jlist_p active_p;

#ifndef PVFS_NR_OPEN
#define PVFS_NR_OPEN (NR_OPEN)
#endif

fdesc_p pfds[PVFS_NR_OPEN] = {NULL};
int pvfs_checks_disabled=0;

static int pvfs_open_called = 0;

static int send_open_req(const char* pathname, char *serverfilename,
	struct sockaddr *saddr, mreq_p);
int unix_open(const char *, int, mode_t, fpart_p);
int pvfs_open(const char*, int, ...);
static int pvfs_open_generic(const char* pathname, int flag, va_list ap);

int pvfs_open64(const char* pathname, int flags, ...)
{
	va_list ap;
	int ret = -1;

	va_start(ap, flags);

	ret = pvfs_open_generic(pathname, flags | O_LARGEFILE, ap);

	va_end(ap);

	return(ret);
}

/* pvfs_open() - PVFS open command.
 * pathname - file to be opened
 * flag - flags (O_RDONLY, etc.) to be used in open
 * possible arguments that get handled by va_arg:
 * ------------------------------------------------
 * mode - if creating file, mode to be used.  Must also be specified
 *	   when providing meta or partation information, though it will be
 *	   ignored unless O_CREAT is set.
 * meta_p - metadata information.  O_META must be set to use this
 * part_p - partition information.  O_PART must be set to use this
 */
int pvfs_open(const char* pathname, int flags, ...)
{
	va_list ap;
	int ret = -1;

	va_start(ap, flags);

	ret = pvfs_open_generic(pathname, flags, ap);

	va_end(ap);

	return(ret);
}


/* pvfs_open_generic()
 *
 * does the real work in open, after the va list and any 64 bit flags
 * have been setup by pvfs_open() or pvfs_open64().
 *
 * returns file descriptor on success, -1 on failure
 */
static int pvfs_open_generic(const char* pathname, int flag, va_list ap)
{
	int 		i, fd, myerr;
	mreq 		req;
	struct sockaddr *saddr;
	char *fn;
	int64_t fs_ino;
	mode_t mode = 0;
	pvfs_filestat* meta_p = NULL;
	fpart* part_p = NULL;

	if (pvfs_open_called == 0) {
		atexit((void *)pvfs_exit);
		pvfs_open_called = 1;
	}

	memset(&req, 0, sizeof(req));

	/* we should look for mode in any of these cases */
	if(flag & O_CREAT || flag & O_META || flag & O_PART)
	{
		mode = va_arg(ap, mode_t);
	}

	/* the meta will be next if it is required */
	if(flag & O_META)
	{
		meta_p = va_arg(ap, pvfs_filestat*);
	}

	/* and finally the partition */
	if(flag & O_PART)
	{
		part_p = va_arg(ap, fpart_p);
	}

	if (!pathname) {
		errno = EFAULT;
		return(-1);
	}
	if (!pvfs_checks_disabled) {
		/* Check to see if file is a PVFS file */
		i = pvfs_detect(pathname, &fn, &saddr, &fs_ino, NULL, FOLLOW_LINK);
		if (i < 0 || i > 2) /* error */ {
			PERROR("Error finding file");
			return(-1);
		}
	}
	else i = 0;

	if (i == 0) /* not a PVFS file or directory */ {
		return(unix_open(pathname, flag, mode, part_p));
	}
	if (i == 2) /* PVFS directory */ {
		if (flag & (O_WRONLY | O_RDWR)) {
			errno = EISDIR;
			return(-1);
		}

		if ((fd = open("/dev/null", O_RDONLY, 0)) < 0) {
			myerr = errno;
			PERROR("pvfs_open: error opening /dev/null instead of metafile");
			errno = myerr;
			return(-1);
		}

		if (!(pfds[fd] = (fdesc_p)malloc(sizeof(fdesc)))) {
			myerr = errno;
			close(fd);
			errno = myerr;
			return(-1);
		}
		memset(pfds[fd], 0, sizeof(fdesc));

		/* save the file name in terms manager would understand */
		if (!(pfds[fd]->fn_p = (char *)malloc(strlen(fn)+1))) {
			myerr = errno;
			free(pfds[fd]);
			pfds[fd] = NULL;
			close(fd);
			errno = myerr;
			return(-1);
		}
		memset(pfds[fd]->fn_p, 0, strlen(fn)+1);
		strcpy(pfds[fd]->fn_p, fn);

#ifdef STRICT_DESC_CHECK
		pfds[fd]->checkval = FDESC_CHECKVAL;
#endif
		pfds[fd]->fs = FS_PDIR;
		pfds[fd]->part_p    = NULL;
		pfds[fd]->dim_p     = NULL;
		pfds[fd]->fd.off    = 0;
		pfds[fd]->fd.grp_nr = -1;
		pfds[fd]->fd.flag   = flag;
		pfds[fd]->fd.cap    = 0;
		pfds[fd]->fd.ref    = 1;

		/* store the manager connection info */
		memcpy(&pfds[fd]->fd.meta.mgr, saddr, sizeof(struct sockaddr));
		return(fd);
	}

	/* otherwise we have a PVFS file... */
	/* Prepare request for file system */
	if (flag & O_META) {
		req.req.open.meta.p_stat.base = meta_p->base;
		req.req.open.meta.p_stat.pcount = meta_p->pcount;
		req.req.open.meta.p_stat.ssize = meta_p->ssize;
	}
	else {
		req.req.open.meta.p_stat.base = -1;
		req.req.open.meta.p_stat.pcount = -1;
		req.req.open.meta.p_stat.ssize = -1;
	}

	req.uid = getuid();
	req.gid = getgid();
	req.type = MGR_OPEN;

	req.dsize = strlen(fn);
	req.req.open.flag = flag;
	req.req.open.mode  = mode;
	req.req.open.meta.fs_ino = fs_ino;
	req.req.open.meta.u_stat.st_uid = req.uid;
	req.req.open.meta.u_stat.st_gid = req.gid;
	req.req.open.meta.u_stat.st_mode = mode;

	if (flag & O_APPEND) 
		req.req.open.flag = flag & ~(O_APPEND);

	if ((fd = send_open_req(pathname, fn, saddr, &req)) < 0) return(-1);

	/* set partition (if provided) */
	if (flag & O_PART) {
		if (!(pfds[fd]->part_p = malloc(sizeof(fpart)))) {
			PERROR("pvfs_open: malloc");
			return(-1);
		}
		*(pfds[fd]->part_p) = *part_p;
	}

	/* open all IOD connections (if requested) */
#ifndef __ALWAYS_CONN__
	if (flag & O_CONN) {
#else
	if (1) {
#endif
		ireq iodreq;

		iodreq.majik_nr   = IOD_MAJIK_NR;
		iodreq.release_nr = PVFS_RELEASE_NR;
		iodreq.type       = IOD_NOOP;
		iodreq.dsize      = 0;

		/* build job to send reqs and recv acks to/from iods */
		if (!active_p) active_p = jlist_new();
		initset(&socks); /* clear out the socket set */

		if (build_simple_jobs(pfds[fd], &iodreq) < 0) {
			ERR("pvfs_open: build_simple_jobs failed...continuing\n");
			return(fd);
		}

		/* call do_job */
		while (!jlist_empty(active_p)) {
			if (do_jobs(active_p, &socks, -1) < 0) {
				ERR("pvfs_open: do_jobs failed...continuing\n");
				return(fd); /* we'll let it slide for now */
			}
		}
		/* don't bother looking for errors */
	}

	/* move to end of file (if necessary) */
	if (flag & O_APPEND) {
		pvfs_lseek(fd, 0, SEEK_END);
	}

	/* tell the sockset library to start with random sockets if enabled */
	randomnextsock(1);
	return(fd);
}

int pvfs_creat(const char* pathname, mode_t mode)
{
	int flags;
	
	/* call pvfs_open with correct flags            */
	flags  =  O_CREAT|O_WRONLY|O_TRUNC;
	return(pvfs_open(pathname, flags, mode, NULL, NULL)); 
}

/* send_open_req() - sends request to the manager, sets up FD structure,
 * returns file descriptor number
 *
 * servername is actually the filename in terms the server can
 * understand, not the name of the server.
 *
 * Returns -1 on error.
 */
static int send_open_req(const char* pathname, char *servername,
	struct sockaddr *saddr, mreq_p req_p)
{
	int fd, s, i;
	mack ack;

	/* Send request to mgr */
	for (;;) {
		s = send_mreq_saddr(saddr, req_p, servername, &ack);
		/* Check ack */
		if (s >= 0 && !ack.status) break; /* went through */

		/* problems... */
		errno = ack.eno;
		return(-1);
	}

	/* always just open /dev/null as a placeholder */
	if ((fd = open("/dev/null", O_RDONLY, 0)) < 0) {
		int myerr = errno;
		PERROR("pvfs_open: error opening /dev/null instead of metafile");
		errno = myerr;
		return(-1);
	}
	LOG1("Opening: %s\n", pathname);

	/* ensure we didn't over-run our PVFS open file table */
	if (fd >= PVFS_NR_OPEN) {
	    ERR("pvfs_open: exceeded max PVFS open files!\n");
	    close(fd);
	    return -1;
	}
	pfds[fd] = (fdesc_p)malloc(sizeof(fdesc) + sizeof(iod_info) * 
		(ack.ack.open.meta.p_stat.pcount - 1));
#ifdef STRICT_DESC_CHECK
	pfds[fd]->checkval  = FDESC_CHECKVAL;
#endif
	pfds[fd]->fs        = FS_PVFS; /* PVFS file type */
	pfds[fd]->fd.off    = 0;
	pfds[fd]->fd.grp_nr = -1;
	pfds[fd]->fd.flag   = req_p->req.open.flag;
	pfds[fd]->fd.meta   = ack.ack.open.meta;
	pfds[fd]->fd.cap    = ack.ack.open.cap;
	pfds[fd]->fd.ref    = 1;
	pfds[fd]->part_p    = NULL;
	pfds[fd]->dim_p     = NULL;
	pfds[fd]->fn_p      = NULL;

   if ((i = brecv(s, pfds[fd]->fd.iod, ack.dsize)) < 0) {
      PERROR("pvfs_open: brecv");
      close(fd);
      return(-1);
   }
	for (i=0; i < pfds[fd]->fd.meta.p_stat.pcount; i++) {
#ifdef STRICT_DESC_CHECK
		pfds[fd]->fd.iod[i].checkval = IOD_INFO_CHECKVAL;
#endif
		/* dont connect to the iod's now. just add them to the iodinfo table */
		if((pfds[fd]->fd.iod[i].sock = instantiate_iod_entry((struct sockaddr *)&pfds[fd]->fd.iod[i].addr)) < 0) {
			int my_err = errno;
			PERROR("Could not instantiate iod entry!\n");
			close(fd);
			errno = my_err;
			return -1;
		}
		inc_ref_count(pfds[fd]->fd.iod[i].sock);
	}
	
	return(fd);
}

int unix_open(const char *pathname, int flag, mode_t mode, fpart_p part_p)
{
	int fd, myerr;
	pvfs_filestat p_stat={0,1,8192,8192}; /* metadata */

	if ((fd = open(pathname, flag & ~PVFSMASK, mode)) < 0) {
		return(fd);
	}
	if (!(pfds[fd] = (fdesc_p)malloc(sizeof(fdesc)))) {
		myerr = errno;
		close(fd);
		errno = myerr;
		return(-1);
	}
#ifdef STRICT_DESC_CHECK
	pfds[fd]->checkval = FDESC_CHECKVAL;
#endif
	pfds[fd]->fs = FS_UNIX;
	pfds[fd]->part_p    = NULL;
	pfds[fd]->dim_p     = NULL;
	pfds[fd]->fn_p      = NULL;
	pfds[fd]->fd.off    = 0;
	pfds[fd]->fd.grp_nr = -1;
	pfds[fd]->fd.flag   = flag;
	pfds[fd]->fd.cap    = 0;
	pfds[fd]->fd.ref    = 1;
	pfds[fd]->fd.meta.p_stat = p_stat; /* fill in some dummy values */
	/* set partition (if provided) */
	if (flag & O_PART) {
		if (!(pfds[fd]->part_p = malloc(sizeof(fpart)))) {
			myerr = errno;
			close(fd);
			errno = myerr;
			return(-1);
		}
		*(pfds[fd]->part_p) = *part_p;
	}
	return(fd);
}

#ifdef STRICT_DESC_CHECK
int do_fdesc_check(int fd)
{
	int i, err=0;
	if (fd < 0 || !pfds[fd]) {
		fprintf(stderr, "do_fdesc_check: invalid fd (fd = %d)\n", fd);
		return(-1);
	}
	if (pfds[fd]->checkval != FDESC_CHECKVAL) {
		fprintf(stderr, "do_fdesc_check: fdesc struct (fd = %d) trashed?\n", fd);
		return(-1);
	}
	if (pfds[fd]->fs != FS_PVFS) return(0);

	for (i=0; i < pfds[fd]->fd.meta.p_stat.pcount; i++) {
		if (pfds[fd]->fd.iod[i].checkval != IOD_INFO_CHECKVAL) {
			err = -1;
			fprintf(stderr, "do_fdesc_check: iod_info struct"
				" (fd = %d, iod = %d) trashed?\n", fd, i);
		}
	}
	return(err);
}

#endif

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */

==> pvfs-1.6.2.tgz: pvfs-1.6.2/lib/pvfs_open.c <==
/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See LIBRARY_COPYING in top-level directory.
 */


/* This file contains PVFS library call for open.   		*/
/* Should work to create and open files 						*/

#include <lib.h>
#include <signal.h>
#include <sockio.h>
#include <sockset.h>
#include <stdarg.h>
#include <errno.h>

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

extern sockset socks;
extern jlist_p active_p;

#ifndef PVFS_NR_OPEN
#define PVFS_NR_OPEN (NR_OPEN)
#endif

fdesc_p pfds[PVFS_NR_OPEN] = {NULL};
int pvfs_checks_disabled=0;

#define DEFAULT_ALLOC 8

static int pvfs_open_called = 0, pvfs_open_optimization = 1, default_alloc = DEFAULT_ALLOC;

static int send_open_req(const char* pathname, char *serverfilename,
	struct sockaddr *saddr, mreq_p);
int unix_open(const char *, int, mode_t, fpart_p);
int pvfs_open(const char*, int, ...);
static int pvfs_open_generic(const char* pathname, int flag, va_list ap);

/*
 * For a given open PVFS file descriptor,
 * return the number of IOD nodes.
 */
static int get_iod_count(int fd)
{
	if (fd < 0 || !pfds[fd]) {
		fprintf(stderr, "get_iod_count: invalid fd (fd = %d)\n", fd);
		return(-1);
	}
#ifdef STRICT_DESC_CHECK
	if (pfds[fd]->checkval != FDESC_CHECKVAL) {
		fprintf(stderr, "get_iod_count: fdesc struct (fd = %d) trashed?\n", fd);
		return(-1);
	}
#endif
	if (pfds[fd]->fs != FS_PVFS) {
		return(-1);
	}
	return pfds[fd]->fd.meta.p_stat.pcount; 
}

int pvfs_open64(const char* pathname, int flags, ...)
{
	va_list ap;
	int ret = -1;

	va_start(ap, flags);

	ret = pvfs_open_generic(pathname, flags | O_LARGEFILE, ap);

	va_end(ap);

	return(ret);
}

/* pvfs_open() - PVFS open command.
 * pathname - file to be opened
 * flag - flags (O_RDONLY, etc.) to be used in open
 * possible arguments that get handled by va_arg:
 * ------------------------------------------------
 * mode - if creating file, mode to be used.  Must also be specified
 *	   when providing meta or partation information, though it will be
 *	   ignored unless O_CREAT is set.
 * meta_p - metadata information.  O_META must be set to use this
 * part_p - partition information.  O_PART must be set to use this
 */
int pvfs_open(const char* pathname, int flags, ...)
{
	va_list ap;
	int ret = -1;

	va_start(ap, flags);

	ret = pvfs_open_generic(pathname, flags, ap);

	va_end(ap);

	return(ret);
}


/* pvfs_open_generic()
 *
 * does the real work in open, after the va list and any 64 bit flags
 * have been setup by pvfs_open() or pvfs_open64().
 *
 * returns file descriptor on success, -1 on failure
 */

static int pvfs_open_generic(const char* pathname, int flag, va_list ap)
{
	int 		i, fd, myerr, niodservers = 0;
	mreq 		req;
	struct sockaddr *saddr;
	char *fn;
	int64_t fs_ino;
	mode_t mode = 0;
	pvfs_filestat* meta_p = NULL;
	fpart* part_p = NULL;

	if (pvfs_open_called == 0) {
		/* Disable the open optimization the first time around */
		pvfs_open_optimization = 0;
		atexit((void *)pvfs_exit);
		pvfs_open_called = 1;
	}

	memset(&req, 0, sizeof(req));

	/* we should look for mode in any of these cases */
	if(flag & O_CREAT || flag & O_META || flag & O_PART)
	{
		mode = va_arg(ap, mode_t);
	}

	/* the meta will be next if it is required */
	if(flag & O_META)
	{
		meta_p = va_arg(ap, pvfs_filestat*);
	}

	/* and finally the partition */
	if(flag & O_PART)
	{
		part_p = va_arg(ap, fpart_p);
	}

#ifdef LARGE_FILE_SUPPORT
	flag |= O_LARGEFILE;
#endif

	if (!pathname) {
		errno = EFAULT;
		return(-1);
	}
	if (!pvfs_checks_disabled) {
		/* Check to see if file is a PVFS file */
		i = pvfs_detect(pathname, &fn, &saddr, &fs_ino, NULL, FOLLOW_LINK);
		if (i < 0 || i > 2) /* error */ {
			PERROR("Error finding file");
			return(-1);
		}
	}
	else i = 0;

	if (i == 0) /* not a PVFS file or directory */ {
		return(unix_open(pathname, flag, mode, part_p));
	}
	if (i == 2) /* PVFS directory */ {
		if (flag & (O_WRONLY | O_RDWR)) {
			errno = EISDIR;
			return(-1);
		}

		if ((fd = open("/dev/null", O_RDONLY, 0)) < 0) {
			myerr = errno;
			PERROR("pvfs_open: error opening /dev/null instead of metafile");
			errno = myerr;
			return(-1);
		}

		if (!(pfds[fd] = (fdesc_p)malloc(sizeof(fdesc)))) {
			myerr = errno;
			close(fd);
			errno = myerr;
			return(-1);
		}
		memset(pfds[fd], 0, sizeof(fdesc));

		/* save the file name in terms manager would understand */
		if (!(pfds[fd]->fn_p = (char *)malloc(strlen(fn)+1))) {
			myerr = errno;
			free(pfds[fd]);
			pfds[fd] = NULL;
			close(fd);
			errno = myerr;
			return(-1);
		}
		memset(pfds[fd]->fn_p, 0, strlen(fn)+1);
		strcpy(pfds[fd]->fn_p, fn);

#ifdef STRICT_DESC_CHECK
		pfds[fd]->checkval = FDESC_CHECKVAL;
#endif
		pfds[fd]->fs = FS_PDIR;
		pfds[fd]->part_p    = NULL;
		pfds[fd]->dim_p     = NULL;
		pfds[fd]->fd.off    = 0;
		pfds[fd]->fd.grp_nr = -1;
		pfds[fd]->fd.flag   = flag;
		pfds[fd]->fd.cap    = 0;
		pfds[fd]->fd.ref    = 1;

		/* store the manager connection info */
		memcpy(&pfds[fd]->fd.meta.mgr, saddr, sizeof(struct sockaddr));
		return(fd);
	}

	/* otherwise we have a PVFS file... */
	/* Prepare request for file system */
	if (flag & O_META) {
		req.req.open.meta.p_stat.base = meta_p->base;
		niodservers = req.req.open.meta.p_stat.pcount = meta_p->pcount;
		req.req.open.meta.p_stat.ssize = meta_p->ssize;
	}
	else {
		req.req.open.meta.p_stat.base = -1;
		req.req.open.meta.p_stat.pcount = -1;
		req.req.open.meta.p_stat.ssize = -1;
		/*
		 * default_alloc is initialized to a default-value which
		 * would most likely be set to the total number of I/O servers
		 * in the cluster. It is changed after the very first open
		 */
		niodservers = pvfs_open_optimization ? default_alloc : 0;
	}

	req.uid = getuid();
	req.gid = getgid();
	req.type = MGR_OPEN;

	req.dsize = strlen(fn);
	req.req.open.flag = flag;
	req.req.open.mode  = mode;
	req.req.open.meta.fs_ino = fs_ino;
	req.req.open.meta.u_stat.st_uid = req.uid;
	req.req.open.meta.u_stat.st_gid = req.gid;
	req.req.open.meta.u_stat.st_mode = mode;
	/* 
	 * A good approximation of how much to expect
	 * is the sizeof(iod_info) * # of IOD servers
	 */
	req.req.open.ackdsize = sizeof(iod_info) * niodservers;

	if (flag & O_APPEND) 
		req.req.open.flag = flag & ~(O_APPEND);

	if ((fd = send_open_req(pathname, fn, saddr, &req)) < 0) {
		return(-1);
	}
	/* set partition (if provided) */
	if (flag & O_PART) {
		if (!(pfds[fd]->part_p = malloc(sizeof(fpart)))) {
			PERROR("pvfs_open: malloc");
			return(-1);
		}
		*(pfds[fd]->part_p) = *part_p;
	}

	/* open all IOD connections (if requested) */
#ifndef __ALWAYS_CONN__
	if (flag & O_CONN) {
#else
	if (1) {
#endif
		ireq iodreq;

		iodreq.majik_nr   = IOD_MAJIK_NR;
		iodreq.release_nr = PVFS_RELEASE_NR;
		iodreq.type       = IOD_NOOP;
		iodreq.dsize      = 0;

		/* build job to send reqs and recv acks to/from iods */
		if (!active_p) active_p = jlist_new();
		initset(&socks); /* clear out the socket set */

		if (build_simple_jobs(pfds[fd], &iodreq) < 0) {
			ERR("pvfs_open: build_simple_jobs failed...continuing\n");
			return(fd);
		}

		/* call do_job */
		while (!jlist_empty(active_p)) {
			if (do_jobs(active_p, &socks, -1) < 0) {
				ERR("pvfs_open: do_jobs failed...continuing\n");
				return(fd); /* we'll let it slide for now */
			}
		}
		/* don't bother looking for errors */
	}

	/* move to end of file (if necessary) */
	if (flag & O_APPEND) {
		pvfs_lseek(fd, 0, SEEK_END);
	}

	/* tell the sockset library to start with random sockets if enabled */
	randomnextsock(1);
	/* Reenable the open optimization if it is the first successful call to pvfs_open */
	if (!pvfs_open_optimization) {
		int tot_iod_count;

		pvfs_open_optimization = 1;
		/* this should return the number of iod's participating */
		if ((tot_iod_count = get_iod_count(fd)) >= 0) { 
			default_alloc = tot_iod_count; 
		}
	}
	/* 
	 * We may want to change the value of default_alloc
	 * after every successful invocation of pvfs_open().
	 */

	/*
	 * if ((tot_iod_count = get_iod_count(fd)) >= 0) { 
	 *		default_alloc = tot_iod_count; 
	 *	}
	 */
	return(fd);
}

int pvfs_creat(const char* pathname, mode_t mode)
{
	int flags;
	
	/* call pvfs_open with correct flags            */
	flags  =  O_CREAT|O_WRONLY|O_TRUNC;
	return(pvfs_open(pathname, flags, mode, NULL, NULL)); 
}

/* send_open_req() - sends request to the manager, sets up FD structure,
 * returns file descriptor number
 *
 * servername is actually the filename in terms the server can
 * understand, not the name of the server.
 *
 * Returns -1 on error.
 */
#define MAX_ACKDSIZE 16384
static int send_open_req(const char* pathname, char *servername,
	struct sockaddr *saddr, mreq_p req_p)
{
	int fd, s, i, pcount = 0;
	mack ack;
	void *ptr = NULL;

	/* Complain if ack.dsize is not a multiple of sizeof(iod_info) */
	if(req_p->req.open.ackdsize % sizeof(iod_info) != 0 
			|| req_p->req.open.ackdsize < 0 
			|| req_p->req.open.ackdsize > MAX_ACKDSIZE) {
		errno = EINVAL;
		PERROR("pvfs_open: Invalid value of ackdsize");
		return -1;
	}

	/* always just open /dev/null as a placeholder */
	if ((fd = open("/dev/null", O_RDONLY, 0)) < 0) {
		int myerr = errno;
		PERROR("pvfs_open: error opening /dev/null instead of metafile");
		errno = myerr;
		return(-1);
	}
	LOG1("Opening: %s\n", pathname);

	/* ensure we didn't over-run our PVFS open file table */
	if (fd >= PVFS_NR_OPEN) {
	    ERR("pvfs_open: exceeded max PVFS open files!\n");
	    close(fd);
	    return -1;
	}
	/* Allocate now, if we are trying to be aggressive */
	if(req_p->req.open.ackdsize > 0) {
		/* guesstimate the number of iod servers */
		pcount = req_p->req.open.ackdsize / sizeof(iod_info);
		pfds[fd] = (fdesc_p)malloc(sizeof(fdesc) + sizeof(iod_info) * 
			(pcount - 1));
		if(!pfds[fd]) {
			close(fd);
			errno = ENOMEM;
			return -1;
		}
		ptr = &pfds[fd]->fd.iod;
	}
	/* Send request to mgr */
	for (;;) {
		memset(&ack, 0, sizeof(ack));
		s = send_mreq_saddr(saddr, req_p, servername, &ack, ptr);
		/* Check ack */
		if (s >= 0 && !ack.status) break; /* went through */
		/* problems... */
		close(fd);
		free(pfds[fd]);
		pfds[fd] = NULL;
		errno = ack.eno;
		return(-1);
	}
	/* Allocate now, if we are trying to be conservative */
	if(req_p->req.open.ackdsize == 0) {
		pfds[fd] = (fdesc_p)malloc(sizeof(fdesc) + sizeof(iod_info) * 
			(ack.ack.open.meta.p_stat.pcount - 1));
		if(!pfds[fd]) {
			close(fd);
			errno = ENOMEM;
			return -1;
		}
	}
#ifdef STRICT_DESC_CHECK
	pfds[fd]->checkval  = FDESC_CHECKVAL;
#endif
	pfds[fd]->fs        = FS_PVFS; /* PVFS file type */
	pfds[fd]->fd.off    = 0;
	pfds[fd]->fd.grp_nr = -1;
	pfds[fd]->fd.flag   = req_p->req.open.flag;
	pfds[fd]->fd.meta   = ack.ack.open.meta;
	pfds[fd]->fd.cap    = ack.ack.open.cap;
	pfds[fd]->fd.ref    = 1;
	pfds[fd]->part_p    = NULL;
	pfds[fd]->dim_p     = NULL;
	pfds[fd]->fn_p      = NULL;

	/* have we received the iod information yet? */
	if(req_p->req.open.ackdsize == 0) {
		/* if not, then we do so now.. */
		if ((i = brecv(s, pfds[fd]->fd.iod, ack.dsize)) < 0) {
			int err = errno;
			PERROR("pvfs_open: brecv");
			close(fd);
			free(pfds[fd]);
			pfds[fd]=NULL;
			errno = err;
			return(-1);
		}
	}
	else {
		/* we have received some portion of it.. but we may have to receive some more */
		if(ack.dsize > req_p->req.open.ackdsize) {
			int64_t remaining = (ack.dsize - req_p->req.open.ackdsize);
			/* We have to call realloc on pfsd[fd] */
			pfds[fd] = realloc(pfds[fd], ack.dsize + sizeof(fdesc));
			if(!pfds[fd]) {
				int err = errno;
				close(fd);
				errno = err;
				return -1;
			}
			/* Now get the remaining */
			if ((i = brecv(s, &pfds[fd]->fd.iod[pcount], remaining)) < 0) {
				int err = errno;
				PERROR("pvfs_open: brecv");
				close(fd);
				free(pfds[fd]);
				pfds[fd]=NULL;
				errno = err;
				return(-1);
			}
		}
		/* otherwise we may have overestimated the amount of memory. but for now
		 * we just leave it and dont free it. FIXME: We may wish to do that though!
		 */ 
	}
	for (i=0; i < pfds[fd]->fd.meta.p_stat.pcount; i++) {
#ifdef STRICT_DESC_CHECK
		pfds[fd]->fd.iod[i].checkval = IOD_INFO_CHECKVAL;
#endif
		/* dont connect to the iod's now. just add them to the iodinfo table */
		if((pfds[fd]->fd.iod[i].sock = instantiate_iod_entry((struct sockaddr *)&pfds[fd]->fd.iod[i].addr)) < 0) {
			int my_err = errno;
			PERROR("Could not instantiate iod entry!\n");
			close(fd);
			errno = my_err;
			return -1;
		}
		inc_ref_count(pfds[fd]->fd.iod[i].sock);
	}
	return(fd);
}

int unix_open(const char *pathname, int flag, mode_t mode, fpart_p part_p)
{
	int fd, myerr;
	pvfs_filestat p_stat={0,1,8192,8192}; /* metadata */

	if ((fd = open(pathname, flag & ~PVFSMASK, mode)) < 0) {
		return(fd);
	}
	if (!(pfds[fd] = (fdesc_p)malloc(sizeof(fdesc)))) {
		myerr = errno;
		close(fd);
		errno = myerr;
		return(-1);
	}
#ifdef STRICT_DESC_CHECK
	pfds[fd]->checkval = FDESC_CHECKVAL;
#endif
	pfds[fd]->fs = FS_UNIX;
	pfds[fd]->part_p    = NULL;
	pfds[fd]->dim_p     = NULL;
	pfds[fd]->fn_p      = NULL;
	pfds[fd]->fd.off    = 0;
	pfds[fd]->fd.grp_nr = -1;
	pfds[fd]->fd.flag   = flag;
	pfds[fd]->fd.cap    = 0;
	pfds[fd]->fd.ref    = 1;
	pfds[fd]->fd.meta.p_stat = p_stat; /* fill in some dummy values */
	/* set partition (if provided) */
	if (flag & O_PART) {
		if (!(pfds[fd]->part_p = malloc(sizeof(fpart)))) {
			myerr = errno;
			close(fd);
			errno = myerr;
			return(-1);
		}
		*(pfds[fd]->part_p) = *part_p;
	}
	return(fd);
}

#ifdef STRICT_DESC_CHECK
int do_fdesc_check(int fd)
{
	int i, err=0;
	if (fd < 0 || !pfds[fd]) {
		fprintf(stderr, "do_fdesc_check: invalid fd (fd = %d)\n", fd);
		return(-1);
	}
	if (pfds[fd]->checkval != FDESC_CHECKVAL) {
		fprintf(stderr, "do_fdesc_check: fdesc struct (fd = %d) trashed?\n", fd);
		return(-1);
	}
	if (pfds[fd]->fs != FS_PVFS) return(0);

	for (i=0; i < pfds[fd]->fd.meta.p_stat.pcount; i++) {
		if (pfds[fd]->fd.iod[i].checkval != IOD_INFO_CHECKVAL) {
			err = -1;
			fprintf(stderr, "do_fdesc_check: iod_info struct"
				" (fd = %d, iod = %d) trashed?\n", fd, i);
		}
	}
	return(err);
}

#endif

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */
